#include "Arduino.h"
#include "LEDStar.h"

const int USB_TIMEOUT = 2000; // Time to wait until serial connection to PC over USB established
const int CMD_PAUSE = 6; // Time to wait after command. Can't be less than 5ms
const byte MODE_IMG = 0; // All calls to setPixel are set in local buffer and sent out only at update() call
const byte MODE_PXL = 1; // All calls to setPixel are sent out immediately

LEDStar::LEDStar()
{
  _mode = MODE_IMG;
}

void LEDStar::init(bool debugMode)
{
  _debugMode = debugMode;
  if (_debugMode) {
    Serial.begin(9600);
    int startTime = millis();
    while (!Serial) {
      // wait for serial port to connect.
      if (millis() - startTime > USB_TIMEOUT) {
        _debugMode = false;
        return;
      }
    }
    _debugLabel("ON");
  }

  Serial1.begin(115200);
  while (!Serial1) {
    ; // wait for serial port to connect.
  }
  Serial1.write("INIT\n");
  delay(CMD_PAUSE);
}

void LEDStar::setMode(byte mode)
{
  _mode = mode;
  _wipeUpdateBuffer();
}

void LEDStar::test()
{
  Serial1.write("TEST\n");
  delay(CMD_PAUSE);
}

void LEDStar::testImg(byte index)
{
  Serial1.write("IMG B ");
  Serial1.write(index);// gap
  Serial1.write(3);// pixel count
  Serial1.write(255); Serial1.write(0); Serial1.write(0);// rgb
  Serial1.write(0); Serial1.write(255); Serial1.write(0);// rgb
  Serial1.write(0); Serial1.write(0); Serial1.write(255);// rgb
  Serial1.write(0);
  Serial1.write(0);
  delay(CMD_PAUSE);
}

void LEDStar::clear()
{
  Serial1.write("CLR\n");
  delay(CMD_PAUSE);
}

void LEDStar::show()
{
  Serial1.write("SHW\n");
  delay(CMD_PAUSE);
}

void LEDStar::setPixel(byte index, byte r, byte g, byte b)
{
  int dataIndex = index * 3;
  _pixelData[dataIndex++] = r;
  _pixelData[dataIndex++] = g;
  _pixelData[dataIndex] = b;
  _pixelUpdate[index] = 1;

  if (_mode == MODE_PXL) {
    Serial1.write("SET ");
    Serial1.print(String(index, HEX));
    Serial1.write(" 1 ");
    Serial1.write(r);//R
    Serial1.write(g);//G
    Serial1.write(b);//B
    Serial1.write("\n");
    delay(CMD_PAUSE);
  }
}

void LEDStar::update()
{
  if (_mode != MODE_IMG) return;
  
  _debugLabel("IMG start");
  Serial1.write("IMG B ");
  byte head = 0;
  byte tail = 0;
  while (true) {
    // Seek for the next updated pixel
    while (head < PIXELS && !_pixelUpdate[head]) {
      head++;
    }
    if (head == PIXELS) {
      // Reached the end
      Serial1.write(0);
      Serial1.write(0);
      delay(CMD_PAUSE);
      _debugLabel("IMG end");
      _wipeUpdateBuffer();
      return;
    }
    Serial1.write(head - tail);// gap
    _debugByte("gap", head - tail);
    tail = head;
    // Seek for the end of pixel block
    while (head < PIXELS && _pixelUpdate[head]) {
      head++;
    }
    Serial1.write(head - tail);// pixel block length
    _debugByte("pxl blk", head - tail);
    int index = 0;
    // Output the pixel data
    for (index = tail * 3; index < head * 3; index++) {
      _debugByte("rgb", _pixelData[index]);
      Serial1.write(_pixelData[index]);
    }
    tail = head;
  }
}

void LEDStar::_wipeUpdateBuffer()
{
  // Wipe the pixel update buffer
  for (byte index = 0; index < PIXELS; index++) {
    _pixelUpdate[index] = 0;
  }
}

void LEDStar::_debugLabel(String label)
{
  if (!_debugMode) return;
  Serial.print("DBG ");
  Serial.println(label);
}

void LEDStar::_debugByte(String label, byte value)
{
  if (!_debugMode) return;
  Serial.print("DBG ");
  Serial.print(label);
  Serial.print(": 0x");
  Serial.println(String(value, HEX));
}

